__all__ = [
    "imshow",
    "ndplot",
    "relplot",
    "roishow",
    "set_style",
]
from magnify.plot.image import roishow, imshow
from magnify.plot.ndplot import ndplot
from magnify.plot.relation import relplot
from magnify.plot.style import set_style

set_style()



__version__ = "0.7.5"

__all__ = []

from magnify.file import *
from magnify.registry import *
import magnify.filter
import magnify.find
import magnify.identify
import magnify.postprocess
import magnify.preprocess
import magnify.reader
import magnify.stitch



from __future__ import annotations

import cv2 as cv
import numpy as np
import xarray as xr

import magnify.registry as registry
import magnify.utils as utils


@registry.component("drop")
def drop(
    assay: xr.Dataset,
    squeeze: bool = True,
    roi_only: bool = False,
    drop_tiles: bool = True,
):
    if squeeze:
        assay = assay.squeeze(drop=True)

    if roi_only:
        return assay.roi
    elif drop_tiles:
        return assay.drop_vars("tile")
    else:
        return assay


@registry.component("circularize")
def circularize(
    assay: xr.Dataset,
):
    subassay = assay.isel(time=0, channel=0)
    fg = subassay.fg.to_numpy()
    bg = subassay.bg.to_numpy()
    for i in range(assay.sizes["mark"]):
        old_fg = utils.to_uint8(fg[i])
        contour = cv.findContours(old_fg, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE)[0][0]
        center, radius = cv.minEnclosingCircle(contour)
        fg[i] = utils.circle(
            fg.shape[-1],
            row=round(center[1]),
            col=round(center[0]),
            radius=round(radius),
            value=True,
        )
        bg[i] = ~fg[i] & bg[i]
    assay.fg[:] = fg[:, np.newaxis, np.newaxis, :, :]
    assay.bg[:] = bg[:, np.newaxis, np.newaxis, :, :]
    return assay



import logging

import matplotlib.pyplot as plt

log_level = logging.DEBUG
text_logger = logging.getLogger("magnify")
title_logger = []
plot_logger = []

def debug(text, data=None):
    if log_level > logging.DEBUG:
        return

    if data is None:
        text_logger.debug(text)
    else:
        title_logger.append(text)
        plot_logger.append(data.copy())

def warning(text, data=None):
    if log_level > logging.WARNING:
        return

    if data is None:
        text_logger.debug(text)
    else:
        title_logger.append(text)
        plot_logger.append(data.copy())

def figure():
    num_plots = len(plot_logger)
    fig, axs = plt.subplots(nrows=num_plots, squeeze=False, figsize=(10, 10 * num_plots))
    axs = axs[:, 0]

    for ax, title, data in zip(axs, title_logger, plot_logger):
        ax.set_title(title)
        ax.imshow(data)

    return fig



from __future__ import annotations
from collections.abc import Callable, Iterable
from typing import Any
import inspect
import re

import cv2 as cv
import numba
import numpy as np
import xarray as xr


def to_uint8(arr: np.ndarray) -> np.ndarray:
    arr = arr.astype(float)
    arr = arr - np.min(arr)
    if np.max(arr) > 0:
        arr = 255 * arr / np.max(arr)
    return arr.astype(np.uint8)


def circle(
    image_length: int, row: int, col: int, radius: int, value: Any = 1, thickness: int = -1
) -> np.ndarray:
    image = np.zeros((image_length, image_length), dtype=np.uint8)
    cv.circle(image, (col, row), radius, 1, thickness=thickness)
    image = image.astype(type(value)) * value
    return image


def annulus(
    image_length: int, row: int, col: int, outer_radius: int, inner_radius: int, value: Any = 1.0
) -> np.ndarray:
    outer_circle = circle(image_length, row, col, outer_radius, value)
    inner_circle = circle(image_length, row, col, inner_radius, value)
    return outer_circle & ~inner_circle


def contour_center(contour):
    m = cv.moments(contour)
    return m["m01"] / m["m00"], m["m10"] / m["m00"]


@numba.jit(nopython=True)
def ceildiv(a: int, b: int) -> int:
    return -(a // -b)


@numba.jit(nopython=True)
def bounding_box(
    x: int, y: int, box_length: int, image_width: int, image_height: int
) -> tuple[int, int, int, int]:
    top = y - box_length // 2
    bottom = y + ceildiv(box_length, 2)
    if top < 0:
        bottom += -top
        top = 0
    if bottom > image_height:
        top -= bottom - image_height
        bottom = image_height
    left = x - box_length // 2
    right = x + ceildiv(box_length, 2)
    if left < 0:
        right += -left
        left = 0
    if right > image_width:
        left -= right - image_width
        right = image_width
    return top, bottom, left, right


def valid_kwargs(kwargs: dict[str, Any], func: Callable) -> dict[str, Any]:
    args = list(inspect.signature(func).parameters)
    return {k: kwargs[k] for k in kwargs if k in args}


def natural_sort_key(s: str) -> list[str]:
    reg = re.compile("([0-9]+)")
    return [int(text) if text.isdigit() else text.lower() for text in reg.split(s)]


def to_list(x: Any) -> list:
    if x is None:
        return []
    elif not isinstance(x, Iterable) or isinstance(x, str):
        return [x]
    else:
        return list(x)



from __future__ import annotations
from typing import cast
import os

import tifffile
import numpy as np
import xarray as xr

import magnify.registry as registry


@registry.component("flatfield_correct")
def flatfield_correct(assay: xr.Dataset, flatfield=1.0, darkfield=0.0):
    if isinstance(flatfield, os.PathLike):
        with tifffile.TiffFile(os.path.expanduser(flatfield)) as tif:
            flatfield = tif.asarray()
    if isinstance(darkfield, os.PathLike):
        with tifffile.TiffFile(os.path.expanduser(darkfield)) as tif:
            darkfield = tif.asarray()

    dtype = assay.tile.dtype
    tiles = (assay.tile.astype(float) - darkfield).clip(min=0)
    max_val = tiles.max()
    tiles = tiles / flatfield
    tiles = tiles * max_val / tiles.max()
    assay["tile"] = tiles.astype(assay.tile.dtype)
    return assay


@registry.component("horizontal_flip")
def horizontal_flip(assay: xr.Dataset):
    if "image" in assay:
        assay["image"] = assay.image.isel(im_x=slice(None, None, -1))
    else:
        assay["tile"] = assay.tile.isel(tile_x=slice(None, None, -1))
    return assay


@registry.component("vertical_flip")
def vertical_flip(assay: xr.Dataset):
    if "image" in assay:
        assay["image"] = assay.image.isel(im_y=slice(None, None, -1))
    else:
        assay["tile"] = assay.tile.isel(tile_y=slice(None, None, -1))
    return assay



from __future__ import annotations

import cv2 as cv
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import plotly.express as px
import plotly.graph_objects as go
import xarray as xr

from magnify.plot.ndplot import ndplot
import magnify.utils as utils


def roishow(
    xp: xr.Dataset,
    facet_col=None,
    animation_frame=None,
    binary_string=True,
    binary_format="jpeg",
    binary_compression_level=0,
    cmap="viridis",
    zmin=None,
    zmax=None,
    **kwargs,
):
    def imfunc(xp: xr.Dataset, **kwargs):
        img = xp.roi.compute()
        if binary_string:
            img = mpl.colors.Normalize(vmin=zmin, vmax=zmax)(img.to_numpy())
            img = plt.get_cmap(cmap)(img)[:, :, :3]
        fig = px.imshow(img, binary_string=binary_string, binary_format=binary_format)
        contours = get_contours(xp)
        fig.add_trace(
            go.Scatter(
                x=np.concatenate([c[:, 0] for c in contours]),
                y=np.concatenate([c[:, 1] for c in contours]),
                mode="lines",
                showlegend=False,
                line_color="green",
            )
        )
        return fig.data

    fig = ndplot(xp, imfunc, animation_frame=animation_frame, facet_col=facet_col, **kwargs)
    fig.update_layout(width=800, height=800, dragmode="pan")
    return fig


def imshow(
    xp: xr.Dataset,
    facet_col=None,
    animation_frame=None,
    binary_string=True,
    binary_format="jpeg",
    binary_compression_level=0,
    compression_ratio=1,
    contour_type="roi",
    cmap="viridis",
    zmin=None,
    zmax=None,
    **kwargs,
):
    def imfunc(xp: xr.Dataset, **kwargs):
        img = xp.image[..., ::compression_ratio, ::compression_ratio].compute()
        if binary_string:
            img = mpl.colors.Normalize(vmin=zmin, vmax=zmax)(img.to_numpy())
            img = plt.get_cmap(cmap)(img)[:, :, :3]
        fig = px.imshow(img, binary_string=binary_string, binary_format=binary_format)
        if "roi" in xp:
            roi = xp.roi.compute()
            # Initialize image metadata.
            roi_len = roi.sizes["roi_y"] // compression_ratio
            valid_x = []
            valid_y = []
            valid_labels = []
            invalid_x = []
            invalid_y = []
            invalid_labels = []
            for idx, m in roi.groupby("mark"):
                x = m.x.item() / compression_ratio
                y = m.y.item() / compression_ratio
                # Get the centers and the bounds of the bounding box.
                top, bottom, left, right = utils.bounding_box(
                    x, y, roi_len, img.shape[1], img.shape[0]
                )
                # Contours are either roi bounding boxes or contours around the foreground.
                if contour_type == "roi":
                    contour_x = [left, left, right, right, left, None]
                    contour_y = [bottom, top, top, bottom, bottom, None]
                elif contour_type == "fg":
                    cont = get_contours(m)
                    if len(cont) == 0:
                        continue
                    # Adjust contours to be in image coordinates.
                    contour_x = list(np.concatenate([c[:, 0] + left for c in cont])) + [None]
                    contour_y = list(np.concatenate([c[:, 1] + top for c in cont])) + [None]

                # Get the label for the bounding box.
                if "tag" in m.coords:
                    label = f"{idx}: {m.tag.item()}"
                else:
                    label = str(idx)
                if m.valid.item():
                    valid_x += contour_x
                    valid_y += contour_y
                    valid_labels += [label] * len(contour_x)
                else:
                    invalid_x += contour_x
                    invalid_y += contour_y
                    invalid_labels += [label] * len(contour_y)

            fig.add_trace(
                go.Scatter(
                    x=valid_x,
                    y=valid_y,
                    mode="lines",
                    hovertemplate="%{text}<extra></extra>",
                    text=valid_labels,
                    showlegend=False,
                    line_color="green",
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=invalid_x,
                    y=invalid_y,
                    mode="lines",
                    hovertemplate="%{text}<extra></extra>",
                    text=invalid_labels,
                    showlegend=False,
                    line_color="red",
                )
            )
        return fig.data

    fig = ndplot(xp, imfunc, animation_frame=animation_frame, facet_col=facet_col, **kwargs)
    fig.update_layout(width=800, height=800, dragmode="pan")
    return fig


def get_contours(roi):
    contours, _ = cv.findContours(
        roi.fg.to_numpy().astype("uint8"),
        cv.RETR_EXTERNAL,
        cv.CHAIN_APPROX_SIMPLE,
    )
    # Remove the extra dimension inserted by opencv.
    contours = [c[:, 0] for c in contours]
    # Close the curves.
    contours = [np.append(c, [c[0]], axis=0) for c in contours]
    return contours



from __future__ import annotations

import cv2 as cv
import numpy as np
import xarray as xr

from magnify import utils
import magnify.registry as registry


@registry.component("filter_expression")
def filter_expression(
    assay: xr.Dataset,
    search_channel: str | list[str] | None = None,
    min_contrast: int | None = None,
):
    if search_channel is None:
        search_channels = assay.channel
    else:
        search_channels = utils.to_list(search_channel)

    valid = xr.zeros_like(assay.valid, dtype=bool)
    for channel in search_channels:
        subassay = assay.isel(time=0).sel(channel=channel)
        fg = subassay.roi.where(subassay.fg).median(dim=["roi_x", "roi_y"]).compute()
        bg = subassay.roi.where(subassay.bg).median(dim=["roi_x", "roi_y"]).compute()
        if min_contrast is None:
            # Compute the intensity differences between every pair of backgrounds on the first timestep.
            bg_n = bg.to_numpy().flatten()
            diffs = bg_n[:, np.newaxis] - bg_n[np.newaxis, :]
            offdiag = np.ones_like(diffs, dtype=bool) & (~np.eye(len(diffs), dtype=bool))
            diffs = diffs[offdiag]

            # Include any markers where the fg - bg is above 5 sigma of the mean difference.
            upper_bound = 4 * diffs.std()
        else:
            upper_bound = min_contrast
        valid |= fg - bg > upper_bound

    assay["valid"] &= valid
    return assay


@registry.component("filter_nonround")
def filter_nonround(
    assay: xr.Dataset, min_roundness: float = 0.75, search_channel: str | list[str] | None = None
):
    if search_channel is None:
        search_channels = assay.channel
    else:
        search_channels = utils.to_list(search_channel)

    valid = assay.valid.to_numpy()
    for channel in search_channels:
        subassay = assay.isel(time=0).sel(channel=channel)
        fg = utils.to_uint8(subassay.fg.to_numpy())
        areas = subassay.fg.sum(dim=["roi_x", "roi_y"]).to_numpy()
        for i in range(assay.sizes["mark"]):
            contours, _ = cv.findContours(fg[i], cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE)
            perimeter = sum(cv.arcLength(c, True) for c in contours)
            area = sum(cv.contourArea(c) for c in contours)
            if perimeter == 0:
                assay.valid[i] = False
                continue
            roundness = 4 * np.pi * float(areas[i]) / perimeter**2
            valid[i] &= roundness > min_roundness
    assay.valid.data = valid

    return assay


@registry.component("filter_leaky")
def filter_leaky_buttons(assay: xr.Dataset, search_channel: str | list[str] | None = None):
    if search_channel is None:
        search_channels = assay.channel
    else:
        search_channels = utils.to_list(search_channel)

    tag = assay.tag.to_numpy()
    valid = assay.valid.to_numpy()
    rows = assay.mark_row.to_numpy()
    for channel in search_channels:
        sub_roi = assay.sel(channel=channel, time=0).roi.compute()
        # Compute the intensity differences between every pair of backgrounds on the first timestep.
        bg = sub_roi.where(sub_roi.bg).median(dim=["roi_x", "roi_y"]).compute()
        bg_n = bg.to_numpy().flatten()
        diffs = bg_n[:, np.newaxis] - bg_n[np.newaxis, :]
        offdiag = np.ones_like(diffs, dtype=bool) & (~np.eye(len(diffs), dtype=bool))
        diffs = diffs[offdiag]

        # Find markers where the fg - bg is below 5 sigma of the mean difference.
        upper_bound = 5 * diffs.std()
        fg = sub_roi.where(sub_roi.fg).median(dim=["roi_x", "roi_y"])
        empty = (fg - bg < upper_bound).to_numpy()
        for i in range(assay.sizes["mark"]):
            row = rows[i]
            if tag[i] == "":
                continue
            if row > 0 and tag[i - 1] == "":
                valid[i] &= empty[i - 1]
            if row < assay.mark_row.max() and tag[i + 1] == "":
                valid[i] &= empty[i + 1]
    assay.valid.data = valid

    return assay



from __future__ import annotations
import logging
from typing import Callable, Sequence

import confection
from numpy.typing import ArrayLike
import xarray as xr

import magnify.logger as logger
import magnify.registry as registry
import magnify.utils as utils


class Pipeline:
    def __init__(self, reader: str, config: dict[str, dict[str, str]]):
        self.reader: Callable[[ArrayLike | str], xr.Dataset] = registry.readers.get(reader)()
        self.config = confection.Config(config)
        self.components: list[Callable[[xr.Dataset], xr.Dataset]] = []

        if "debug" in self.config and self.config["debug"] == True:
            logger.log_level = logging.DEBUG
        else:
            logger.log_level = logging.INFO

    def __call__(
        self,
        data: ArrayLike | str,
        times: Sequence[int] | None = None,
        channels: Sequence[str] | None = None,
    ) -> xr.Dataset | list[xr.Dataset]:
        inputs = self.reader(data=data, times=times, channels=channels)
        assays = []
        for assay in inputs:
            for name, component in self.components:
                assay = component(assay)

            assays.append(assay)

        if len(assays) == 1:
            assays = assays[0]

        return assays

    def add_pipe(
        self,
        name: str,
        after: str | int | None = None,
        before: str | int | None = None,
        first: bool = False,
        last: bool = False,
    ) -> None:
        component_factory = registry.components.get(name)
        component = component_factory(**utils.valid_kwargs(self.config, component_factory))

        if after is None and before is None and not first and not last:
            last = True
        if (after is not None) + (before is not None) + first + last > 1:
            raise ValueError("Only one of after, before, first, and last can be set.")
        new_element = (name, component)
        if first:
            idx = 0
        elif last:
            idx = len(self.components)
        elif isinstance(before, int):
            idx = before
        elif isinstance(before, str):
            idx = [name for (name, _) in self.components].index(before)
        elif isinstance(after, int):
            idx = after + 1
        elif isinstance(after, str):
            idx = [name for (name, _) in self.components].index(after) + 1

        self.components.insert(idx, new_element)

    def add_pipe(
        self,
        name: str,
        after: str | int | None = None,
        before: str | int | None = None,
        first: bool = False,
        last: bool = False,
    ) -> None:
        component_factory = registry.components.get(name)
        component = component_factory(**utils.valid_kwargs(self.config, component_factory))

        if after is None and before is None and not first and not last:
            last = True
        if (after is not None) + (before is not None) + first + last > 1:
            raise ValueError("Only one of after, before, first, and last can be set.")
        new_element = (name, component)
        if first:
            idx = 0
        elif last:
            idx = len(self.components)
        elif isinstance(before, int):
            idx = before
        elif isinstance(before, str):
            idx = [name for (name, _) in self.components].index(before)
        elif isinstance(after, int):
            idx = after + 1
        elif isinstance(after, str):
            idx = [name for (name, _) in self.components].index(after) + 1

        self.components.insert(idx, new_element)

    def remove_pipe(self, name: str) -> None:
        idx = list(zip(*self.components))[0].index(name)
        self.components.pop(idx)
